import numpy as np
import scipy.constants
import typing


from spdm.utils.typing import array_type
from spdm.utils.tags import _not_found_
from spdm.core.expression import Expression, Variable, smooth, zero
from spdm.core.sp_property import sp_tree
from fytok.modules.core_profiles import CoreProfiles, CoreProfilesSpecies
from fytok.modules.equilibrium import Equilibrium
from fytok.modules.core_sources import CoreSources
from fytok.utils.atoms import atoms
from fytok.utils.logger import logger


@sp_tree
class CollisionalEquipartition(CoreSources.Source):
    identifier = "collisional_equipartition"

    code = {"name": "collisional_equipartition", "description": "Fusion reaction"}  # type: ignore

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D) -> CoreSources.Source.TimeSlice:
        ii_collision: bool = self.code.parameters.ii_collision
        ie_collision: bool = self.code.parameters.ie_collision
        current: CoreSources.Source.TimeSlice = super().fetch(profiles_1d)

        source_1d = current.profiles_1d

        e = scipy.constants.elementary_charge
        ze = -1.0
        me = scipy.constants.electron_mass
        ae = scipy.constants.electron_mass / scipy.constants.atomic_mass

        ne = profiles_1d.electrons.density
        Te = profiles_1d.electrons.temperature
        ve = profiles_1d.electrons.velocity.toroidal

        conductivity_parallel = zero

        source_1d.electrons.energy = zero

        species: typing.List[CoreProfilesSpecies] = [*profiles_1d.ion]

        for i, ion_i in enumerate(species):
            zi = ion_i.z
            ai = ion_i.a
            mi = ion_i.a * scipy.constants.atomic_mass

            ni = ion_i.density
            Ti = ion_i.temperature
            vi = ion_i.velocity.toroidal

            if Ti is _not_found_:
                continue

            # clog_ei = piecewise(
            #     [
            #         (
            #             23 - np.log(zi) - 0.5 * np.log(ne * 1.0e-6) + 1.5 * np.log(Te),
            #             ((Ti * me / mi) < Te) & (Te <= (10 * zi * zi)),
            #         ),
            #         (
            #             24 - 0.5 * np.log(ne * 1.0e-6) + np.log(Te),
            #             (((Ti * me / mi) < (10 * zi * zi)) & ((10 * zi * zi) < Te)),
            #         ),
            #         (
            #             16 - np.log(zi * zi * ae) - 0.5 * np.log(ne * 1.0e-6) + 1.5 * np.log(Te),
            #             Te <= (Ti * me / mi),
            #         ),
            #     ],
            #     name="clog",
            #     label=r"\Lambda_{ei}",
            # )
            delta0 = np.heaviside(Te - Ti * me / mi, 0.5)
            delta1 = np.heaviside(Te - 10 * zi * zi, 0.5)
            delta2 = np.heaviside(Ti * me / mi - 10 * zi * zi, 0.5)

            clog_ei = (
                +(23 - np.log(zi) - 0.5 * np.log(ne * 1.0e-6) + 1.5 * np.log(Te)) * delta0 * (1 - delta1)
                + (24 - 0.5 * np.log(ne * 1.0e-6) + np.log(Te)) * delta1
                + (16 - np.log(zi * zi * ai) - 0.5 * np.log(ni * 1.0e-6) + 1.5 * np.log(Ti)) * (1 - delta0)
            )

            clog_ei._metadata["name"] = "clog"
            clog_ei._metadata["label"] = r"\Lambda_{ei}"

            nv_ei = (
                (3.2e-9 * zi * zi / ai) * (Te ** (-1.5)) * clog_ei * 1.5e-6
            )  # FIXME: 1.5e-6 符合 astra 结果，需要再次复核

            conductivity_parallel += 1.96e-15 * e**2 / me * ne * ne / nv_ei

            nv_ei = ni * ne * nv_ei

            Tie = Ti - Te

            # energy exchange term
            source_1d.ion[ion_i.label].energy -= Tie * nv_ei
            source_1d.electrons.energy += Tie * nv_ei

            # momentum exchange term
            if vi is not _not_found_ and ve is not _not_found_:
                source_1d.ion[ion_i.label].momentum.toroidal -= (vi - ve) * nv_ei
                source_1d.electrons.momentum.toroidal += (ve - vi) * nv_ei

            # collisions frequency and energy exchange terms:
            # @ref NRL 2019 p.34
            for ion_j in species[i + 1 :]:
                if ion_i.label == ion_j.label:
                    continue
                # ion-Ion collisions:

                zj = ion_j.z
                aj = ion_j.a
                mj = ion_j.a * scipy.constants.atomic_mass

                nj = ion_j.density
                Tj = ion_j.temperature
                vj = ion_j.velocity.toroidal

                if Tj is _not_found_:
                    continue

                # Coulomb logarithm:
                clog = (
                    (23 - np.log(zi * zj * (ai + aj)))
                    + np.log(Ti * aj + Tj * ai)
                    - 0.5 * np.log((ni * zi * zi / Ti + nj * zj * zj / Tj) * 1.0e-6)
                )

                nv_ij = (
                    1.8e-19
                    * np.sqrt(mi * mj * 1.0e-6)
                    * (zi * zi * zj * zj)
                    * (((Ti * mj + Tj * mi) * 1.0e-3) ** (-1.5))
                    * clog
                    * 1.0e-6
                )

                # nv_ij = smooth(nv_ij, window_length=3, polyorder=2)

                nv_ij = ni * nj * nv_ij

                Tij = Ti - Tj

                # energy exchange term
                source_1d.ion[ion_i.label].energy -= Tij * nv_ij
                source_1d.ion[ion_j.label].energy += Tij * nv_ij

                # momentum exchange term
                if vi is not _not_found_ and vj is not _not_found_:
                    source_1d.ion[ion_i.label].momentum.toroidal += (vi - vj) * nv_ij
                    source_1d.ion[ion_j.label].momentum.toroidal += (vj - vi) * nv_ij

                # Tij = Expression(deburr, Ti - Tj)

                ##############################
                # 增加阻尼，消除震荡
                # epsilon = 1.0e-10
                # c = (1.5 - 1.0 / (1.0 + np.exp(-np.abs(Ti - Tj) / (Ti + Tj) / epsilon))) * 2
                # Tij = (Ti - Tj) * c
                # if isinstance(c, array_type):
                #     logger.debug((c,))
                ##############################
        # Plasma electrical conductivity:
        source_1d.conductivity_parallel = conductivity_parallel
        return current


CoreSources.Source.register("collisional_equipartition", CollisionalEquipartition)



import scipy.constants
import numpy as np

from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_sources import CoreSources
from fytok.modules.equilibrium import Equilibrium
from fytok.utils.atoms import atoms
from fytok.utils.logger import logger

from spdm.core.function import Function
from spdm.core.expression import Variable, Expression, piecewise
from spdm.core.sp_property import sp_tree
from spdm.utils.typing import array_type


@CoreSources.Source.register(["bootstrap"])
@sp_tree
class BootstrapCurrent(CoreSources.Source):
    identifier = "bootstrap_current"
    code = {
        "name": "bootstrap_current",
        "description": "Bootstrap current, based on  Tokamaks, 3ed, sec 14.12 J.A.Wesson 2003",
    }

    def fetch(self, x: Variable | array_type, **variables: Expression) -> CoreSources.Source.TimeSlice:
        current: CoreSources.Source.TimeSlice = super().fetch(x, **variables)
        equilibrium: Equilibrium.TimeSlice = self.inputs.get_source("equilibrium").time_slice.current

        source_1d = current.profiles_1d
        eq_1d = equilibrium.profiles_1d

        grid = source_1d.grid

        rho_tor_boundary = grid.rho_tor_boundary

        rho_tor = rho_tor_boundary * x

        psi_norm = Function(grid.rho_tor_norm, grid.psi_norm, label=r"\bar{\psi}")(x)

        eV = scipy.constants.electron_volt

        B0 = equilibrium.vacuum_toroidal_field.b0
        R0 = equilibrium.vacuum_toroidal_field.r0

        q = eq_1d.q(psi_norm)

        # max(np.asarray(1.07e-4*((Te[0]/1000)**(1/2))/B0), rho_tor[0])   # Larmor radius,   eq 14.7.2

        Te = variables.get("electrons/temperature")
        ne = variables.get("electrons/density")
        Pe = scipy.constants.k / scipy.constants.electron_volt * Te * ne
        dlnTe = Te.dln
        dlnne = ne.dln
        dlnPe = dlnne + dlnTe

        # Coulomb logarithm
        #  Ch.14.5 p727 Tokamaks 2003
        lnCoul = piecewise(
            [
                ((14.9 - 0.5 * np.log(ne / 1e20) + np.log(Te / 1000)), (Te < 10)),
                ((15.2 - 0.5 * np.log(ne / 1e20) + np.log(Te / 1000)), (Te >= 10)),
            ],
            name="clog",
            label=r"\Lambda_{e}",
        )

        # electron collision time , eq 14.6.1
        tau_e = 1.09e16 * ((Te / 1000.0) ** (3 / 2)) / ne / lnCoul

        vTe = np.sqrt(Te * eV / scipy.constants.electron_mass)

        epsilon = rho_tor / R0
        epsilon12 = np.sqrt(epsilon)
        epsilon32 = epsilon12**3

        nu_e = R0 * q / vTe / tau_e / epsilon32
        # Zeff = core_profile.zeff

        s = eq_1d.trapped_fraction(psi_norm)  # np.sqrt(2*epsilon)  #
        c1 = (4.0 + 2.6 * s) / (1.0 + 1.02 * np.sqrt(nu_e) + 1.07 * nu_e) / (1.0 + 1.07 * epsilon32 * nu_e)
        c3 = (7.0 + 6.5 * s) / (1.0 + 0.57 * np.sqrt(nu_e) + 0.61 * nu_e) / (1.0 + 0.61 * epsilon32 * nu_e) - c1 * 5 / 2

        j_bootstrap = c1 * dlnPe + c3 * dlnTe

        for k, ni in variables.items():
            k_ = k.split("/")
            if not (k_[0] == "ion" and k_[-1] == "density"):
                continue

            s = k[1]
            Ti = variables[f"ion/{s}/temperature"]

            dlnTi = Ti.dln
            dlnNi = ni.dln
            dlnPi = dlnNi + dlnTi
            mi = atoms[s].mass

            # ion collision time Tokamaks 3ed, eq 14.6.2 p730
            tau_i = 6.6e17 * np.sqrt(mi) * ((Ti / 1000) ** (3 / 2)) / ni / (1.1 * lnCoul)

            # thermal velocity
            v_Ti = np.sqrt(Ti * (eV / scipy.constants.m_p / mi))

            nu_i = R0 * q / epsilon32 / v_Ti / tau_i

            #########################################################################
            #  Sec 14.12 Bootstrap current

            c2 = c1 * Ti / Te

            e3n2 = (epsilon**3) * (nu_i**2)

            c4 = (
                ((-1.17 / (1.0 + 0.46 * x) + 0.35 * np.sqrt(nu_i)) / (1 + 0.7 * np.sqrt(nu_i)) + 2.1 * e3n2)
                / (1 - e3n2)
                / (1 + e3n2)
                * c2
            )

            j_bootstrap += c2 * dlnPi + c4 * dlnTi
            # eq 4.9.2
            # j_bootstrap = j_bootstrap + Ni*Ti*eV*(2.44*dlnne - 0.42*dlnTi)
            #########################################################################

        # eq 4.9.2
        # src.j_bootstrap = (-(q/B0/epsilon12))*j_bootstrap
        fpol = eq_1d.f(psi_norm)
        j_bootstrap = (
            -j_bootstrap
            * x
            / (2.4 + 5.4 * x + 2.6 * x**2)
            * Pe
            * fpol
            * q
            / x
            / (rho_tor[-1]) ** 2
            / (2.0 * scipy.constants.pi * B0)
        )

        source_1d["j_bootstrap"] = j_bootstrap

        return current



import os
import pathlib

import numpy as np
from fytok.tokamak import Tokamak
from fytok.utils.logger import logger
from spdm.core.file import File
from spdm.core.entry import Entry, open_entry
from spdm.view.sp_view import display

WORKSPACE = "/home/salmon/workspace"  # "/ssd01/salmon_work/workspace/"

os.environ["SP_DATA_MAPPING_PATH"] = f"{WORKSPACE}/fytok_data/mapping"

if __name__ == "__main__":

    output_path = pathlib.Path(f"{WORKSPACE}/output")

    device = "iter"

    # entry = open_entry(f"file+mhdin://{WORKSPACE}/omas/omas/machine_mappings/support_files/{device}/04202005Av1_0/mhdin.dat")

    # with File(f"/home/salmon/workspace/fytok_data/mapping/{device}/imas/3/{device}.xml", mode="w", root="mapping", format="xml") as f:
    #     f.write(entry.dump())

    # entry = open_entry(f"file+mhdin://{WORKSPACE}/omas/omas/machine_mappings/support_files/iter/mhdin.dat")

    # with File(f"/home/salmon/workspace/fytok_data/mapping/iter_n/imas/3/iter_n.xml", mode="w", root="mapping", format="xml") as f:
    #     f.write(entry.dump())

    tok = Tokamak(device)

    display(tok, title=f"{tok.device.upper()} RZ   View ", output=output_path/f"{tok.device}_rz.svg")
    # display(tok, title=f"{tok.device.upper()} Top  View ", output=output_path/"east_top.svg", view="TOP")



import pathlib
import re
import scipy.constants
import numpy as np
import pandas as pd

from spdm.core.expression import  Variable
from spdm.core.file import File
from spdm.core.entry import Entry
from spdm.utils.typing import _not_found_
from spdm.numlib.smooth import smooth_1d

PI = scipy.constants.pi
TWOPI = scipy.constants.pi * 2.0


def step_function(x, scale=1.0e-2):
    return 1 / (1 + np.exp(-x / scale))


def read_iter_profiles(path):
    path = pathlib.Path(path)

    excel_file = pd.read_excel(path, sheet_name=1)

    entry = Entry(
        {
            "dataset_fair": {
                "identifier": "15MA Inductive at burn-ASTRA",
                "provenance": {"node": [{"path": "core_profiles", "sources": f"{path.as_posix()}"}]},
            }
        }
    )

    profiles_0D = {}

    for s in excel_file.iloc[0, 3:7]:
        res = re.match(r"(\w+)=(\d+\.?\d*)(\D+)", s)
        profiles_0D[res.group(1)] = (float(res.group(2)), str(res.group(3)))

    profiles_1D = pd.read_excel(path, sheet_name=1, header=10, usecols="B:BN")

    time = 0.0

    R0 = profiles_0D["R"][0]
    B0 = profiles_0D["B"][0]

    vacuum_toroidal_field = {"r0": R0, "b0": B0}

    rho_tor_norm = profiles_1D["x"].values
    rho_tor = profiles_1D["rho"].values
    psi_norm = profiles_1D["Fp"].values

    grid = {
        "rho_tor_norm": rho_tor_norm,
        "rho_tor_boundary": rho_tor[-1],
        "psi_norm": psi_norm,
        "psi_boundary": _not_found_,
        "psi_axis": _not_found_,
    }

    entry["core_profiles"] = {"time_slice": [{"time": time, "vacuum_toroidal_field": vacuum_toroidal_field}]}

    # Core profile
    r_ped = 0.96  # np.sqrt(0.88)
    i_ped = np.argmin(np.abs(rho_tor_norm - r_ped))
    # fmt:off
    psi_norm = profiles_1D["Fp"].values
    # bs_psi = bs_psi_norm*(psi_boundary-psi_axis)+psi_axis

    b_Te =    smooth_1d(rho_tor_norm, profiles_1D["TE"].values,      i_end=i_ped-10, window_len=21)*1000
    b_Ti =    smooth_1d(rho_tor_norm, profiles_1D["TI"].values,      i_end=i_ped-10, window_len=21)*1000
    b_ne =    smooth_1d(rho_tor_norm, profiles_1D["NE"].values,      i_end=i_ped-10, window_len=21)*1.0e19
    b_nDT =   smooth_1d(rho_tor_norm, profiles_1D["Nd+t"].values,    i_end=i_ped-10, window_len=21)*1.0e19*0.5
    b_nath =   smooth_1d(rho_tor_norm, profiles_1D["Nath"].values,    i_end=i_ped-10, window_len=21)*1.0e19
    b_nalpha =   smooth_1d(rho_tor_norm, profiles_1D["Nalf"].values,    i_end=i_ped-10, window_len=21)*1.0e19
    b_nImp =  smooth_1d(rho_tor_norm, profiles_1D["Nz"].values,      i_end=i_ped-10, window_len=21)*1.0e19
    b_zeff = profiles_1D["Zeff"].values
    # fmt:on

    z_eff_star = b_zeff - (b_nDT * 2.0 + 4 * b_nath) / b_ne
    z_imp = 1 - (b_nDT * 2.0 + 2 * b_nath) / b_ne
    b = -2 * z_imp / (0.02 + 0.0012)
    c = (z_imp**2 - 0.02 * z_eff_star) / 0.0012 / (0.02 + 0.0012)

    z_Ar = np.asarray((-b + np.sqrt(b**2 - 4 * c)) / 2)
    z_Be = np.asarray((z_imp - 0.0012 * z_Ar) / 0.02)
    # b_nDT = b_ne * (1.0 - 0.02*4 - 0.0012*18) - b_nHe*2.0

    # Zeff = Function(bs_r_norm, baseline["Zeff"].values)
    # e_parallel = baseline["U"].values / (TWOPI * R0)

    entry["core_profiles/time_slice/0/profiles_1d"] = {
        "time": 0.0,
        "grid": grid,
        "electrons": {"label": "e", "density": b_ne, "temperature": b_Te},
        "ion": [
            {"@name": "D", "density": b_nDT, "temperature": b_Ti},
            {"@name": "T", "density": b_nDT, "temperature": b_Ti},
            {"@name": "Be", "density": 0.02 * b_ne, "z_ion_1d": z_Be},
            {"@name": "Ar", "density": 0.0012 * b_ne, "z_ion_1d": z_Ar},
            {"@name": "He", "density": b_nath, "temperature": b_Ti},
            {"@name": "alpha", "density": b_nalpha - b_nath},
        ],
        # "e_field": {"parallel":  Function(e_parallel,bs_r_norm)},
        # "conductivity_parallel": Function(baseline["Joh"].values*1.0e6 / baseline["U"].values * (TWOPI * grid.r0),bs_r_norm),
        "rho_tor": profiles_1D["rho"].values,
        "q": profiles_1D["q"].values,
        "zeff": profiles_1D["Zeff"].values,
        "vloop": profiles_1D["U"].values,
        "j_ohmic": profiles_1D["Joh"].values * 1.0e6,
        "j_non_inductive": profiles_1D["Jnoh"].values * 1.0e6,
        "j_bootstrap": profiles_1D["Jbs"].values * 1.0e6,
        "j_total": profiles_1D["Jtot"].values * 1.0e6,
        "XiNC": profiles_1D["XiNC"].values,
        "ffprime": profiles_1D["EQFF"].values,
        "pprime": profiles_1D["EQPF"].values,
    }
    _x = Variable(0, name="rho_tor_norm", label=r"\bar{\rho}_{tor}")

    # entry["core_transport"] = {
    #     "model": [
    #         {
    #             "code": {"name": "dummy"},
    #             "time_slice": [
    #                 {
    #                     "time": time,
    #                     "vacuum_toroidal_field": vacuum_toroidal_field,
    #                 }
    #             ],
    #         }
    #     ]
    # }
    # # Core profiles
    # r_ped = 0.96  # np.sqrt(0.88)
    # i_ped = np.argmin(np.abs(rho_tor_norm - r_ped))
    # # Core Transport
    # Cped = 0.17
    # Ccore = 0.4
    # # Function( profiles["Xi"].values,bs_r_norm)  Cped = 0.2
    # # chi = Piecewise(
    # #     [
    # #         (Ccore * (1.0 + 3 * (_x**2)), (_x < r_ped)),
    # #         (Cped, (_x >= r_ped)),
    # #     ],
    # #     label=r"\chi",
    # # )
    # # chi_e = Piecewise(
    # #     [
    # #         (0.5 * Ccore * (1.0 + 3 * (_x**2)), (_x < r_ped)),
    # #         (Cped, (_x >= r_ped)),
    # #     ],
    # #     label=r"\chi_e",
    # # )
    # delta = step_function_approx(_x - r_ped, scale=0.005)
    # chi = (Ccore * (1.0 + 3 * (_x**2))) * (1 - delta) + Cped * delta
    # chi_e = (0.5 * Ccore * (1.0 + 3 * (_x**2))) * (1 - delta) + Cped * delta
    # D = 0.1 * (chi + chi_e)
    # v_pinch_ne = 0.6 * D * _x / R0
    # v_pinch_Te = 2.5 * chi_e * _x / R0
    # v_pinch_ni = D * _x / R0
    # v_pinch_Ti = chi * _x / R0
    # entry["core_transport/model/0/time_slice/0/flux_multiplier"] = 3 / 2
    # entry["core_transport/model/0/time_slice/0/profiles_1d"] = {
    #     "grid_d": grid,
    #     "electrons": {"@name": "e", "particles": {"d": D, "v": v_pinch_ne}, "energy": {"d": chi_e, "v": v_pinch_Te}},
    #     "ion": [
    #         {"@name": "D", "particles": {"d": D, "v": v_pinch_ni}, "energy": {"d": chi, "v": v_pinch_Ti}},
    #         {"@name": "T", "particles": {"d": D, "v": v_pinch_ni}, "energy": {"d": chi, "v": v_pinch_Ti}},
    #         {"@name": "He", "particles": {"d": D, "v": v_pinch_ni}, "energy": {"d": chi, "v": v_pinch_Ti}},
    #         {"@name": "alpha", "particles": {"d": 0.001 * D, "v": 0}},
    #     ],
    # }

    entry["core_sources"] = {
        "source": [
            {"code": {"name": "dummy"}, "time_slice": [{"time": time, "vacuum_toroidal_field": vacuum_toroidal_field}]}
        ]
    }

    S = 9e20 * np.exp(15.0 * (_x**2 - 1.0))

    Q_e = (
        (
            # profiles_1D["Poh"].values
            +profiles_1D["Paux"].values
            - profiles_1D["Prad"].values
            - profiles_1D["Pneu"].values
            # - profiles_1D["Peic"].values
            # + profiles_1D["Pdte"].values
        )
        * 1e6
        / scipy.constants.electron_volt
    )

    Q_DT = (
        (
            +profiles_1D["Pibm"].values
            # + profiles_1D["Peic"].values
            # + profiles_1D["Pdti"].values
        )
        * 1e6
        / scipy.constants.electron_volt
    )

    # Q_He = (profiles_1D["Pdti"].values + profiles_1D["Pdte"].values) * 1e6 / scipy.constants.electron_volt

    # Core Source
    entry["core_sources/source/0/time_slice/0/profiles_1d"] = {
        "grid": grid,
        "conductivity_parallel": profiles_1D["Joh"].values * 1.0e6 / profiles_1D["U"].values * (TWOPI * R0),
        "j_parallel": -profiles_1D["Jtot"].values * 1e6,  # A/m^2
        "electrons": {"@name": "e", "particles": S, "energy": Q_e},
        "ion": [
            {"@name": "D", "particles": S * 0.5, "energy": Q_DT * 0.5},
            {"@name": "T", "particles": S * 0.5, "energy": Q_DT * 0.5},
            # {"@name": "He", "particles": S * 0.00, "energy": Q_DT * 0.00},  #
            # {"@name": "alpha", "particles": S * 0.2, "energy": Q_DT * 0.01},
        ],
    }

    return entry


@File.register(["iterprofiles"])
class ITERProfiles(File):
    """Read iter_profiles.xslx file"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self) -> Entry:
        if self.url.authority:
            raise NotImplementedError(f"{self.url}")

        return read_iter_profiles(pathlib.Path(self.url.path))

    def write(self, d, *args, **kwargs):
        raise NotImplementedError(f"TODO: write ITERProfiles {self.url}")
